# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add required flags to output gcloud run deploy command."""

import os
from typing import Any, Mapping, Sequence
from googlecloudsdk.command_lib.app.gae_to_cr_migration_util.common import util


def translate_add_required_flags(
    input_data: Mapping[str, any],
    source_path: str,
    runtime_base_image: str | None,
) -> Sequence[str]:
  """Add required flags to gcloud run deploy command.

  Args:
    input_data: A mapping containing the translated data from app.yaml.
    source_path: The path to the application source code.
    runtime_base_image: The base image to use for the runtime, returned by
      export_image_api response (only for image based migration).

  Returns:
    A sequence of strings representing the required flags.
  """
  is_flex = util.is_flex_env(input_data)
  migration_tool = (
      'gcloud-app-migrate-flexible-v1'
      if is_flex
      else 'gcloud-app-migrate-standard-v1'
  )
  required_flags = [
      f'--labels={_get_labels(migration_tool=migration_tool)}'
  ]

  if not is_flex:
    base_image = runtime_base_image or input_data['runtime']
    if source_path and _check_dockerfile_exists(source_path):
      required_flags.extend([
          '--clear-base-image',
      ])
    else:
      if base_image:
        required_flags.append(f'--base-image={base_image}')
  required_flags.append('--no-cpu-throttling')

  return required_flags


def update_service_yaml_with_required_flags(
    service_yaml: dict[str, Any],
    input_data: Mapping[str, Any],
) -> None:
  """Updates the service_yaml dict with required labels and annotations.

  Args:
    service_yaml: A dictionary representing the Cloud Run service.yaml.
    input_data: A mapping containing the translated data from app.yaml.
  """
  is_flex = util.is_flex_env(input_data)
  migration_tool = (
      'gcloud-app-migrate-flexible-v1'
      if is_flex
      else 'gcloud-app-migrate-standard-v1'
  )

  # Update labels
  metadata = service_yaml.setdefault('metadata', {})
  labels = metadata.setdefault('labels', {})
  labels['migrated-from'] = 'app-engine'
  labels['migration-tool'] = migration_tool

  # Update annotations
  spec = service_yaml.setdefault('spec', {})
  template = spec.setdefault('template', {})
  template_metadata = template.setdefault('metadata', {})
  annotations = template_metadata.setdefault('annotations', {})
  annotations['run.googleapis.com/cpu-throttling'] = 'false'


def _get_labels(*, migration_tool: str) -> str:
  """Get labels for gcloud run deploy command."""
  return f'migrated-from=app-engine,migration-tool={migration_tool}'


def _check_dockerfile_exists(source_path: str) -> bool:
  """Check for a Dockerfile in the source directory.

  This function verifies if a Dockerfile exists in the same directory
  as the provided `source_path` (typically the app.yaml file).

  Args:
    source_path: The path to the application source code (e.g., app.yaml).

  Returns:
    True if a Dockerfile exists in the same directory as `source_path`, False
    otherwise.
  """
  dockerfile_path = os.path.join(
      os.path.dirname(source_path), 'Dockerfile'
  )
  return os.path.exists(dockerfile_path)


def update_service_yaml_with_base_image(
    service_yaml: dict[str, Any],
    runtime_base_image: str | None,
) -> None:
  """Updates the service_yaml dict with base image annotations if provided.

  Args:
    service_yaml: A dictionary representing the Cloud Run service.yaml.
    runtime_base_image: The base image to use for the runtime.
  """
  if not runtime_base_image:
    return

  spec = service_yaml.setdefault('spec', {})
  template = spec.setdefault('template', {})

  # Set template metadata annotations
  template_metadata = template.setdefault('metadata', {})
  annotations = template_metadata.setdefault('annotations', {})
  annotations['run.googleapis.com/base-images'] = (
      f'{{"app":"{runtime_base_image}"}}'
  )

  # Set template spec runtimeClassName
  template_spec = template.setdefault('spec', {})
  template_spec['runtimeClassName'] = (
      'run.googleapis.com/linux-base-image-update'
  )

  # Ensure container name is set to 'app'
  containers = template_spec.setdefault('containers', [{}])
  if containers:
    containers[0]['name'] = 'app'



# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for creating instance templates running Docker images."""


from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute import containers_utils
from googlecloudsdk.api_lib.compute import image_utils
from googlecloudsdk.api_lib.compute import instance_template_utils
from googlecloudsdk.api_lib.compute import instance_utils
from googlecloudsdk.api_lib.compute import metadata_utils
from googlecloudsdk.api_lib.compute import utils
from googlecloudsdk.api_lib.compute.instances.create import utils as create_utils
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.compute import completers
from googlecloudsdk.command_lib.compute import flags
from googlecloudsdk.command_lib.compute.instance_templates import flags as instance_templates_flags
from googlecloudsdk.command_lib.compute.instances import flags as instances_flags
from googlecloudsdk.command_lib.compute.resource_policies import flags as maintenance_flags
from googlecloudsdk.command_lib.util.args import labels_util
from googlecloudsdk.core import log


def _Args(
    parser,
    release_track,
    container_mount_enabled=False,
    enable_guest_accelerators=False,
    support_multi_writer=True,
    support_region_instance_template=False,
    support_specific_then_x_affinity=False,
    support_any_reservation_then_fail_affinity=False,
    support_disk_labels=False,
    support_ipv6_only=False,
    support_flex_start=False,
    support_skip_guest_os_shutdown=False,
    support_workload_identity_config=False,
    support_vsock_mode=False,
):
  """Add flags shared by all release tracks."""
  parser.display_info.AddFormat(instance_templates_flags.DEFAULT_LIST_FORMAT)
  metadata_utils.AddMetadataArgs(parser)
  instances_flags.AddDiskArgs(
      parser, container_mount_enabled=container_mount_enabled
  )
  instances_flags.AddCreateDiskArgs(
      parser,
      container_mount_enabled=container_mount_enabled,
      support_multi_writer=support_multi_writer,
      support_disk_labels=support_disk_labels,
  )
  if release_track == base.ReleaseTrack.ALPHA:
    instances_flags.AddLocalSsdArgsWithSize(parser)
  elif release_track == base.ReleaseTrack.BETA:
    instances_flags.AddLocalSsdArgs(parser)
  instances_flags.AddCanIpForwardArgs(parser)
  instances_flags.AddContainerMountDiskFlag(parser)
  instances_flags.AddAddressArgs(
      parser,
      instances=False,
      containers=True,
      support_ipv6_only=support_ipv6_only,
  )
  instances_flags.AddMachineTypeArgs(parser)
  deprecate_maintenance_policy = release_track in [base.ReleaseTrack.ALPHA]
  instances_flags.AddMaintenancePolicyArgs(parser, deprecate_maintenance_policy)
  instances_flags.AddNoRestartOnFailureArgs(parser)
  instances_flags.AddPreemptibleVmArgs(parser)
  instances_flags.AddServiceAccountAndScopeArgs(parser, False)
  instances_flags.AddTagsArgs(parser)
  instances_flags.AddCustomMachineTypeArgs(parser)
  instances_flags.AddNetworkArgs(parser)
  instances_flags.AddKonletArgs(parser)
  instances_flags.AddImageArgs(parser)
  instances_flags.AddMinCpuPlatformArgs(parser, release_track)
  instances_flags.AddNetworkTierArgs(parser, instance=True)
  instances_flags.AddConfidentialComputeArgs(parser)
  instances_flags.AddShieldedInstanceConfigArgs(parser)
  instances_flags.AddIPv6AddressArgs(parser)
  instances_flags.AddIPv6PrefixLengthArgs(parser)
  labels_util.AddCreateLabelsFlags(parser)
  instances_flags.AddPrivateNetworkIpArgs(parser)
  instances_flags.AddStackTypeArgs(parser, support_ipv6_only)
  instances_flags.AddIpv6NetworkTierArgs(parser)
  instances_flags.AddInternalIPv6AddressArgs(parser)
  instances_flags.AddInternalIPv6PrefixLengthArgs(parser)
  maintenance_flags.AddResourcePoliciesArgs(
      parser, 'added to', 'instance-template'
  )
  instances_flags.AddReservationAffinityGroup(
      parser,
      group_text="""Specifies the reservation for instances created from this template.""",
      affinity_text="""The type of reservation for instances created from this template.""",
      support_specific_then_x_affinity=support_specific_then_x_affinity,
      support_any_reservation_then_fail_affinity=support_any_reservation_then_fail_affinity,
  )
  instances_flags.AddProvisioningModelVmArgs(
      parser, support_flex_start=support_flex_start
  )

  if enable_guest_accelerators:
    instances_flags.AddAcceleratorArgs(parser)

  flags.AddRegionFlag(
      parser, resource_type='instance template', operation_type='create'
  )

  if support_region_instance_template:
    parser.add_argument(
        '--subnet-region', help='Specifies the region of the subnetwork.'
    )

  parser.add_argument(
      '--description',
      help='Specifies a textual description for the instance template.',
  )

  CreateWithContainer.InstanceTemplateArg = (
      instance_templates_flags.MakeInstanceTemplateArg()
  )
  CreateWithContainer.InstanceTemplateArg.AddArgument(
      parser, operation_type='create'
  )

  parser.display_info.AddCacheUpdater(completers.InstanceTemplatesCompleter)
  if support_skip_guest_os_shutdown:
    instances_flags.AddSkipGuestOsShutdownArgs(parser)
  if support_workload_identity_config:
    instances_flags.AddWorkloadIdentityConfigArgs(parser)
  if support_vsock_mode:
    instances_flags.AddVsockModeArgs(parser)


@base.Deprecate(
    is_removed=False,
    warning=(
        'The option to deploy a container during VM creation using the'
        ' container startup agent is deprecated. Use alternative services to'
        ' run containers on your VMs. Learn more at'
        ' https://cloud.google.com/compute/docs/containers/migrate-containers.'
    ),
    error=(
        'The option to deploy a container during VM creation using the'
        ' container startup agent is deprecated. Use alternative services to'
        ' run containers on your VMs. Learn more at'
        ' https://cloud.google.com/compute/docs/containers/migrate-containers.'
    ),
)
# TODO(b/305707695):Change @base.DefaultUniverseOnly to
# @base.UniverseCompatible once b/305707695 is fixed.
# See go/gcloud-cli-running-tpc-tests.
@base.DefaultUniverseOnly
@base.ReleaseTracks(base.ReleaseTrack.GA, base.ReleaseTrack.PREVIEW)
class CreateWithContainer(base.CreateCommand):
  """Command for creating VM instance templates hosting Docker images."""

  _support_specific_then_x_affinity = False
  _support_any_reservation_then_fail_affinity = False
  _support_disk_labels = False
  _support_skip_guest_os_shutdown = True
  _support_workload_identity_config = False
  _support_vsock_mode = False

  @staticmethod
  def Args(parser):
    _Args(
        parser,
        base.ReleaseTrack.GA,
        container_mount_enabled=True,
        support_multi_writer=False,
        support_region_instance_template=False,
        support_specific_then_x_affinity=False,
        support_any_reservation_then_fail_affinity=False,
        support_disk_labels=False,
        support_ipv6_only=True,
        support_skip_guest_os_shutdown=True,
        support_workload_identity_config=False,
    )
    instances_flags.AddPrivateIpv6GoogleAccessArgForTemplate(
        parser, utils.COMPUTE_GA_API_VERSION
    )

  def _ValidateArgs(self, args):
    instances_flags.ValidateKonletArgs(args)
    instances_flags.ValidateDiskCommonFlags(args)
    instances_flags.ValidateServiceAccountAndScopeArgs(args)
    instances_flags.ValidateNicFlags(args)
    instances_flags.ValidateReservationAffinityGroup(args)
    if instance_utils.UseExistingBootDisk(args.disk or []):
      raise exceptions.InvalidArgumentException(
          '--disk', 'Boot disk specified for containerized VM.'
      )

  def _GetBootDiskSize(self, args):
    boot_disk_size_gb = utils.BytesToGb(args.boot_disk_size)
    utils.WarnIfDiskSizeIsTooSmall(boot_disk_size_gb, args.boot_disk_type)
    return boot_disk_size_gb

  def _GetInstanceTemplateRef(self, args, holder):
    return CreateWithContainer.InstanceTemplateArg.ResolveAsResource(
        args, holder.resources
    )

  def _GetUserMetadata(
      self,
      args,
      client,
      instance_template_ref,
      container_mount_disk_enabled=False,
      container_mount_disk=None,
  ):
    user_metadata = instance_utils.GetValidatedMetadata(args, client)
    return containers_utils.CreateKonletMetadataMessage(
        client.messages,
        args,
        instance_template_ref.Name(),
        user_metadata,
        container_mount_disk_enabled,
        container_mount_disk,
    )

  def _GetNetworkInterfaces(self, args, client, holder):
    if args.network_interface:
      return instance_template_utils.CreateNetworkInterfaceMessages(
          resources=holder.resources,
          scope_lister=flags.GetDefaultScopeLister(client),
          messages=client.messages,
          network_interface_arg=args.network_interface,
          subnet_region=args.region,
      )
    stack_type = getattr(args, 'stack_type', None)
    ipv6_network_tier = getattr(args, 'ipv6_network_tier', None)
    ipv6_address = getattr(args, 'ipv6_address', None)
    ipv6_prefix_length = getattr(args, 'ipv6_prefix_length', None)
    external_ipv6_address = getattr(args, 'external_ipv6_address', None)
    external_ipv6_prefix_length = getattr(
        args, 'external_ipv6_prefix_length', None
    )
    internal_ipv6_address = getattr(args, 'internal_ipv6_address', None)
    internal_ipv6_prefix_length = getattr(
        args, 'internal_ipv6_prefix_length', None
    )
    return [
        instance_template_utils.CreateNetworkInterfaceMessage(
            resources=holder.resources,
            scope_lister=flags.GetDefaultScopeLister(client),
            messages=client.messages,
            network=args.network,
            private_ip=args.private_network_ip,
            subnet_region=args.region,
            subnet=args.subnet,
            address=(
                instance_template_utils.EPHEMERAL_ADDRESS
                if not args.no_address and not args.address
                else args.address
            ),
            network_tier=getattr(args, 'network_tier', None),
            stack_type=stack_type,
            ipv6_network_tier=ipv6_network_tier,
            ipv6_address=ipv6_address,
            ipv6_prefix_length=ipv6_prefix_length,
            external_ipv6_address=external_ipv6_address,
            external_ipv6_prefix_length=external_ipv6_prefix_length,
            internal_ipv6_address=internal_ipv6_address,
            internal_ipv6_prefix_length=internal_ipv6_prefix_length,
        )
    ]

  def _GetScheduling(self, args, client):
    skip_guest_os_shutdown = None
    if self._support_skip_guest_os_shutdown and args.IsKnownAndSpecified(
        'skip_guest_os_shutdown'
    ):
      skip_guest_os_shutdown = args.skip_guest_os_shutdown
    vsock_mode = None
    if self._support_vsock_mode and args.IsSpecified('vsock_mode'):
      vsock_mode = args.vsock_mode
    return instance_utils.CreateSchedulingMessage(
        messages=client.messages,
        maintenance_policy=args.maintenance_policy,
        preemptible=args.preemptible,
        provisioning_model=args.provisioning_model,
        restart_on_failure=args.restart_on_failure,
        skip_guest_os_shutdown=skip_guest_os_shutdown,
        vsock_mode=vsock_mode,
    )

  def _GetServiceAccounts(self, args, client):
    if args.no_service_account:
      service_account = None
    else:
      service_account = args.service_account
    return instance_utils.CreateServiceAccountMessages(
        messages=client.messages,
        scopes=[] if args.no_scopes else args.scopes,
        service_account=service_account,
    )

  def _GetImageUri(self, args, client, holder, instance_template_ref):
    if (
        args.IsSpecified('image')
        or args.IsSpecified('image_family')
        or args.IsSpecified('image_project')
    ):
      image_expander = image_utils.ImageExpander(client, holder.resources)
      image_uri, _ = image_expander.ExpandImageFlag(
          user_project=instance_template_ref.project,
          image=args.image,
          image_family=args.image_family,
          image_project=args.image_project,
      )
      if holder.resources.Parse(image_uri).project != 'cos-cloud':
        log.warning(
            'This container deployment mechanism requires a '
            'Container-Optimized OS image in order to work. Select an '
            'image from a cos-cloud project (cos-stable, cos-beta, '
            'cos-dev image families).'
        )
    else:
      image_uri = containers_utils.ExpandKonletCosImageFlag(client)
    return image_uri

  def _GetMachineType(self, args):
    return instance_utils.InterpretMachineType(
        machine_type=args.machine_type,
        custom_cpu=args.custom_cpu,
        custom_memory=args.custom_memory,
        ext=getattr(args, 'custom_extensions', None),
        vm_type=getattr(args, 'custom_vm_type', None),
    )

  def _GetDisks(
      self,
      args,
      client,
      holder,
      instance_template_ref,
      image_uri,
      match_container_mount_disks=False,
  ):
    boot_disk_size_gb = self._GetBootDiskSize(args)
    # create boot disk through args.boot_disk_device_name
    create_boot_disk = not (
        instance_utils.UseExistingBootDisk(
            (args.disk or []) + (args.create_disk or [])
        )
    )
    return instance_template_utils.CreateDiskMessages(
        args,
        client,
        holder.resources,
        instance_template_ref.project,
        image_uri,
        boot_disk_size_gb,
        create_boot_disk=create_boot_disk,
        match_container_mount_disks=match_container_mount_disks,
        support_disk_labels=self._support_disk_labels,
    )

  def Run(self, args):
    """Issues an InstanceTemplates.Insert request.

    Args:
      args: the argparse arguments that this command was invoked with.

    Returns:
      an InstanceTemplate message object
    """
    self._ValidateArgs(args)
    instances_flags.ValidateNetworkTierArgs(args)

    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    container_mount_disk = instances_flags.GetValidatedContainerMountDisk(
        holder, args.container_mount_disk, args.disk, args.create_disk
    )

    client = holder.client
    instance_template_ref = self._GetInstanceTemplateRef(args, holder)
    image_uri = self._GetImageUri(args, client, holder, instance_template_ref)
    labels = containers_utils.GetLabelsMessageWithCosVersion(
        None, image_uri, holder.resources, client.messages.InstanceProperties
    )
    argument_labels = labels_util.ParseCreateArgs(
        args, client.messages.InstanceProperties.LabelsValue
    )
    if argument_labels:
      labels.additionalProperties.extend(argument_labels.additionalProperties)
    metadata = self._GetUserMetadata(
        args,
        client,
        instance_template_ref,
        container_mount_disk_enabled=True,
        container_mount_disk=container_mount_disk,
    )
    network_interfaces = self._GetNetworkInterfaces(args, client, holder)
    scheduling = self._GetScheduling(args, client)
    service_accounts = self._GetServiceAccounts(args, client)
    machine_type = self._GetMachineType(args)
    disks = self._GetDisks(
        args,
        client,
        holder,
        instance_template_ref,
        image_uri,
        match_container_mount_disks=True,
    )
    confidential_instance_config = (
        create_utils.BuildConfidentialInstanceConfigMessage(
            messages=client.messages,
            args=args,
            support_confidential_compute_type=True,
            support_confidential_compute_type_tdx=True,
        )
    )
    workload_identity_config = (
        instance_utils.CreateWorkloadIdentityConfigMessage(
            args, client.messages, self._support_workload_identity_config
        )
    )

    properties = client.messages.InstanceProperties(
        machineType=machine_type,
        disks=disks,
        canIpForward=args.can_ip_forward,
        confidentialInstanceConfig=confidential_instance_config,
        labels=labels,
        metadata=metadata,
        minCpuPlatform=args.min_cpu_platform,
        networkInterfaces=network_interfaces,
        serviceAccounts=service_accounts,
        scheduling=scheduling,
        tags=containers_utils.CreateTagsMessage(client.messages, args.tags),
    )

    if self._support_workload_identity_config and workload_identity_config:
      properties.workloadIdentityConfig = workload_identity_config

    if args.resource_policies:
      properties.resourcePolicies = args.resource_policies

    if args.private_ipv6_google_access_type is not None:
      properties.privateIpv6GoogleAccess = (
          instances_flags.GetPrivateIpv6GoogleAccessTypeFlagMapperForTemplate(
              client.messages
          ).GetEnumForChoice(args.private_ipv6_google_access_type)
      )

    shielded_instance_config_message = (
        create_utils.BuildShieldedInstanceConfigMessage(
            messages=client.messages, args=args
        )
    )
    if shielded_instance_config_message:
      properties.shieldedInstanceConfig = shielded_instance_config_message

    properties.reservationAffinity = instance_utils.GetReservationAffinity(
        args,
        client,
        support_specific_then_x_affinity=self._support_specific_then_x_affinity,
        support_any_reservation_then_fail_affinity=self._support_any_reservation_then_fail_affinity,
    )

    request = client.messages.ComputeInstanceTemplatesInsertRequest(
        instanceTemplate=client.messages.InstanceTemplate(
            properties=properties,
            description=args.description,
            name=instance_template_ref.Name(),
        ),
        project=instance_template_ref.project,
    )

    return client.MakeRequests(
        [(client.apitools_client.instanceTemplates, 'Insert', request)]
    )


@base.ReleaseTracks(base.ReleaseTrack.BETA)
class CreateWithContainerBeta(CreateWithContainer):
  """Command for creating VM instance templates hosting Docker images."""

  _support_specific_then_x_affinity = True
  _support_any_reservation_then_fail_affinity = True
  _support_disk_labels = True
  _support_skip_guest_os_shutdown = True
  _support_workload_identity_config = False

  @staticmethod
  def Args(parser):
    _Args(
        parser,
        base.ReleaseTrack.BETA,
        container_mount_enabled=True,
        enable_guest_accelerators=True,
        support_region_instance_template=False,
        support_specific_then_x_affinity=True,
        support_any_reservation_then_fail_affinity=True,
        support_disk_labels=True,
        support_ipv6_only=True,
        support_flex_start=True,
        support_skip_guest_os_shutdown=True,
        support_workload_identity_config=False,
        support_vsock_mode=False,
    )
    instances_flags.AddPrivateIpv6GoogleAccessArgForTemplate(
        parser, utils.COMPUTE_BETA_API_VERSION
    )

  def _ValidateArgs(self, args):
    super(CreateWithContainerBeta, self)._ValidateArgs(args)
    instances_flags.ValidateLocalSsdFlags(args)

  def Run(self, args):
    """Issues an InstanceTemplates.Insert request.

    Args:
      args: the argparse arguments that this command was invoked with.

    Returns:
      an InstanceTemplate message object
    """
    self._ValidateArgs(args)
    instances_flags.ValidateNetworkTierArgs(args)

    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    container_mount_disk = instances_flags.GetValidatedContainerMountDisk(
        holder, args.container_mount_disk, args.disk, args.create_disk
    )
    client = holder.client
    instance_template_ref = self._GetInstanceTemplateRef(args, holder)
    image_uri = self._GetImageUri(args, client, holder, instance_template_ref)
    labels = containers_utils.GetLabelsMessageWithCosVersion(
        None, image_uri, holder.resources, client.messages.InstanceProperties
    )
    argument_labels = labels_util.ParseCreateArgs(
        args, client.messages.InstanceProperties.LabelsValue
    )
    if argument_labels:
      labels.additionalProperties.extend(argument_labels.additionalProperties)

    metadata = self._GetUserMetadata(
        args,
        client,
        instance_template_ref,
        container_mount_disk_enabled=True,
        container_mount_disk=container_mount_disk,
    )
    network_interfaces = self._GetNetworkInterfaces(args, client, holder)
    scheduling = self._GetScheduling(args, client)
    service_accounts = self._GetServiceAccounts(args, client)
    machine_type = self._GetMachineType(args)
    disks = self._GetDisks(
        args,
        client,
        holder,
        instance_template_ref,
        image_uri,
        match_container_mount_disks=True,
    )
    confidential_instance_config = (
        create_utils.BuildConfidentialInstanceConfigMessage(
            messages=client.messages,
            args=args,
            support_confidential_compute_type=True,
            support_confidential_compute_type_tdx=True,
        )
    )
    guest_accelerators = (
        instance_template_utils.CreateAcceleratorConfigMessages(
            client.messages, getattr(args, 'accelerator', None)
        )
    )
    workload_identity_config = (
        instance_utils.CreateWorkloadIdentityConfigMessage(
            args, client.messages, self._support_workload_identity_config
        )
    )

    properties = client.messages.InstanceProperties(
        machineType=machine_type,
        disks=disks,
        canIpForward=args.can_ip_forward,
        confidentialInstanceConfig=confidential_instance_config,
        labels=labels,
        metadata=metadata,
        minCpuPlatform=args.min_cpu_platform,
        networkInterfaces=network_interfaces,
        serviceAccounts=service_accounts,
        scheduling=scheduling,
        tags=containers_utils.CreateTagsMessage(client.messages, args.tags),
        guestAccelerators=guest_accelerators,
    )

    if self._support_workload_identity_config and workload_identity_config:
      properties.workloadIdentityConfig = workload_identity_config

    if args.resource_policies:
      properties.resourcePolicies = args.resource_policies

    if args.private_ipv6_google_access_type is not None:
      properties.privateIpv6GoogleAccess = (
          instances_flags.GetPrivateIpv6GoogleAccessTypeFlagMapperForTemplate(
              client.messages
          ).GetEnumForChoice(args.private_ipv6_google_access_type)
      )

    shielded_instance_config_message = (
        create_utils.BuildShieldedInstanceConfigMessage(
            messages=client.messages, args=args
        )
    )
    if shielded_instance_config_message:
      properties.shieldedInstanceConfig = shielded_instance_config_message

    properties.reservationAffinity = instance_utils.GetReservationAffinity(
        args,
        client,
        support_specific_then_x_affinity=self._support_specific_then_x_affinity,
        support_any_reservation_then_fail_affinity=self._support_any_reservation_then_fail_affinity,
    )

    request = client.messages.ComputeInstanceTemplatesInsertRequest(
        instanceTemplate=client.messages.InstanceTemplate(
            properties=properties,
            description=args.description,
            name=instance_template_ref.Name(),
        ),
        project=instance_template_ref.project,
    )

    return client.MakeRequests(
        [(client.apitools_client.instanceTemplates, 'Insert', request)]
    )


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class CreateWithContainerAlpha(CreateWithContainerBeta):
  """Command for creating VM instance templates hosting Docker images."""

  _support_specific_then_x_affinity = True
  _support_any_reservation_then_fail_affinity = True
  _support_disk_labels = True
  _support_ipv6_only = True
  _support_skip_guest_os_shutdown = True
  _support_workload_identity_config = True
  _support_vsock_mode = True

  @staticmethod
  def Args(parser):
    _Args(
        parser,
        base.ReleaseTrack.ALPHA,
        container_mount_enabled=True,
        enable_guest_accelerators=True,
        support_region_instance_template=True,
        support_specific_then_x_affinity=True,
        support_any_reservation_then_fail_affinity=True,
        support_disk_labels=True,
        support_ipv6_only=True,
        support_flex_start=True,
        support_skip_guest_os_shutdown=True,
        support_workload_identity_config=True,
        support_vsock_mode=True,
    )
    instances_flags.AddLocalNvdimmArgs(parser)
    instances_flags.AddPrivateIpv6GoogleAccessArgForTemplate(
        parser, utils.COMPUTE_ALPHA_API_VERSION
    )
    instances_flags.AddIPv6AddressAlphaArgs(parser)
    instances_flags.AddIPv6PrefixLengthAlphaArgs(parser)

  def Run(self, args):
    """Issues an InstanceTemplates.Insert request.

    Args:
      args: the argparse arguments that this command was invoked with.

    Returns:
      an InstanceTemplate message object
    """
    self._ValidateArgs(args)
    instances_flags.ValidateNetworkTierArgs(args)

    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    container_mount_disk = instances_flags.GetValidatedContainerMountDisk(
        holder, args.container_mount_disk, args.disk, args.create_disk
    )

    client = holder.client
    instance_template_ref = self._GetInstanceTemplateRef(args, holder)
    image_uri = self._GetImageUri(args, client, holder, instance_template_ref)
    labels = containers_utils.GetLabelsMessageWithCosVersion(
        None, image_uri, holder.resources, client.messages.InstanceProperties
    )
    argument_labels = labels_util.ParseCreateArgs(
        args, client.messages.InstanceProperties.LabelsValue
    )
    if argument_labels:
      labels.additionalProperties.extend(argument_labels.additionalProperties)

    metadata = self._GetUserMetadata(
        args,
        client,
        instance_template_ref,
        container_mount_disk_enabled=True,
        container_mount_disk=container_mount_disk,
    )
    network_interfaces = self._GetNetworkInterfaces(args, client, holder)
    scheduling = self._GetScheduling(args, client)
    service_accounts = self._GetServiceAccounts(args, client)
    machine_type = self._GetMachineType(args)
    disks = self._GetDisks(
        args,
        client,
        holder,
        instance_template_ref,
        image_uri,
        match_container_mount_disks=True,
    )
    confidential_instance_config = (
        create_utils.BuildConfidentialInstanceConfigMessage(
            messages=client.messages,
            args=args,
            support_confidential_compute_type=True,
            support_confidential_compute_type_tdx=True,
        )
    )
    guest_accelerators = (
        instance_template_utils.CreateAcceleratorConfigMessages(
            client.messages, getattr(args, 'accelerator', None)
        )
    )
    workload_identity_config = (
        instance_utils.CreateWorkloadIdentityConfigMessage(
            args, client.messages, self._support_workload_identity_config
        )
    )

    properties = client.messages.InstanceProperties(
        machineType=machine_type,
        disks=disks,
        canIpForward=args.can_ip_forward,
        confidentialInstanceConfig=confidential_instance_config,
        labels=labels,
        metadata=metadata,
        minCpuPlatform=args.min_cpu_platform,
        networkInterfaces=network_interfaces,
        serviceAccounts=service_accounts,
        scheduling=scheduling,
        tags=containers_utils.CreateTagsMessage(client.messages, args.tags),
        guestAccelerators=guest_accelerators,
    )

    if self._support_workload_identity_config and workload_identity_config:
      properties.workloadIdentityConfig = workload_identity_config

    if args.resource_policies is not None:
      properties.resourcePolicies = args.resource_policies

    if args.private_ipv6_google_access_type is not None:
      properties.privateIpv6GoogleAccess = (
          instances_flags.GetPrivateIpv6GoogleAccessTypeFlagMapperForTemplate(
              client.messages
          ).GetEnumForChoice(args.private_ipv6_google_access_type)
      )

    shielded_instance_config_message = (
        create_utils.BuildShieldedInstanceConfigMessage(
            messages=client.messages, args=args
        )
    )
    if shielded_instance_config_message:
      properties.shieldedInstanceConfig = shielded_instance_config_message

    properties.reservationAffinity = instance_utils.GetReservationAffinity(
        args,
        client,
        support_specific_then_x_affinity=self._support_specific_then_x_affinity,
        support_any_reservation_then_fail_affinity=self._support_any_reservation_then_fail_affinity,
    )

    request = client.messages.ComputeInstanceTemplatesInsertRequest(
        instanceTemplate=client.messages.InstanceTemplate(
            properties=properties,
            description=args.description,
            name=instance_template_ref.Name(),
        ),
        project=instance_template_ref.project,
    )

    return client.MakeRequests(
        [(client.apitools_client.instanceTemplates, 'Insert', request)]
    )


CreateWithContainer.detailed_help = {
    'brief': """\
    Creates a Compute Engine a virtual machine instance template that runs
    a Docker container.
    """,
    'DESCRIPTION': """\
        *{command}* creates a Compute Engine virtual
        machine instance template that runs a container image. To create
        an instance template named 'instance-template-1' that runs the
        'busybox' image, run:

          $ {command} instance-template-1 \
             --container-image=gcr.io/google-containers/busybox

        For more examples, refer to the *EXAMPLES* section below.
        """,
    'EXAMPLES': """\
        To create a template named 'instance-template-1' that runs the
        gcr.io/google-containers/busybox image and executes 'echo "Hello world"'
        as a command, run:

          $ {command} instance-template-1 \
            --container-image=gcr.io/google-containers/busybox \
            --container-command='echo "Hello world"'

        To create a template running gcr.io/google-containers/busybox in
        privileged mode, run:

          $ {command} instance-template-1 \
            --container-image=gcr.io/google-containers/busybox \
            --container-privileged
        """,
}



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines arguments for BigLake commands."""

from googlecloudsdk.api_lib.biglake import util
from googlecloudsdk.calliope import arg_parsers


def AddDescriptionArg(parser):
  """Adds argument for description."""
  parser.add_argument(
      '--description',
      help='Description of the resource.',
  )


def AddPropertiesArg(parser):
  """Adds argument for creating properties."""
  parser.add_argument(
      '--properties',
      metavar='KEY=VALUE',
      type=arg_parsers.ArgDict(),
      help=(
          'Properties associated with the namespace.'
      ),
  )


def AddUpdatePropertiesArgs(parser):
  """Adds arguments for updating properties."""
  parser.add_argument(
      '--update-properties',
      metavar='KEY=VALUE',
      type=arg_parsers.ArgDict(),
      help=(
          'List of properties to update or add.'
      ),
  )
  parser.add_argument(
      '--remove-properties',
      metavar='KEY',
      type=arg_parsers.ArgList(),
      help=(
          'List of properties to remove.'
      ),
  )
  parser.add_argument(
      '--clear-properties',
      action='store_true',
      default=False,
      help=(
          'Clear all properties from the namespace.'
      ),
  )


def AddTableDescribeArgs(parser):
  """Adds arguments for describing tables."""
  parser.add_argument(
      '--snapshots',
      help='Snapshot to get.',
  )


def AddTableRegisterArgs(parser):
  """Adds arguments for registering tables."""
  parser.add_argument(
      '--metadata-location',
      required=True,
      help='Metadata location of the table.',
  )
  parser.add_argument(
      '--overwrite',
      action='store_true',
      default=False,
      help='Overwrite the table if it already exists.',
  )


def AddCatalogsCreateArgs(parser):
  """Adds arguments for creating catalogs."""
  parser.add_argument(
      '--primary-location',
      help=(
          'Primary location for mirroring the remote catalog metadata. It'
          ' must be a BigLake-supported location, and it should be proximate to'
          ' the remote catalog\'s location for better performance and lower'
          ' cost.'
      ),
  )
  AddDescriptionArg(parser)
  AddRestrictedLocationsArg(parser)
  AddDefaultLocationArg(parser)


def AddServiceDirectoryNameArg(parser):
  """Adds argument for service directory name."""
  parser.add_argument(
      '--service-directory-name',
      help=(
          'The service directory resource name for request routing, in the'
          ' format'
          ' `projects/{project_id}/locations/{location}/namespaces/{namespace}/services/{service_name}`.'
          ' see https://docs.cloud.google.com/lakehouse/docs/set-up-cross-cloud-lakehouse#setup-cci.'
      ),
  )


def AddUnityServicePrincipalApplicationIdArg(parser):
  """Adds argument for service principal application ID."""
  parser.add_argument(
      '--unity-service-principal-application-id',
      help=(
          'Optional. The application ID of the Databricks service principal'
          ' that will be used to access the Unity Catalog in the OIDC'
          ' authentication flow. With OIDC, the secret-name argument is not'
          ' used.'
      ),
  )


def AddTableCreateArgs(parser):
  """Adds arguments for creating tables."""
  parser.add_argument(
      '--create-from-file',
      required=True,
      help=(
          'Path to a JSON file containing the table creation request. The'
          ' format must follow the Apache Iceberg REST Catalog Open API'
          ' specification for `CreateTableRequest`. The fields `name` and'
          ' `schema` must be specified.'
      ),
  )


def AddFederatedCatalogArgs(
    parser, support_snowflake=False, support_workday=False
):
  """Adds arguments for federated catalogs."""
  choices = ['unity', 'glue']
  hidden_choices = []
  if support_snowflake:
    choices.append('snowflake')
    hidden_choices.append('snowflake')
  if support_workday:
    choices.append('workday')
    hidden_choices.append('workday')

  parser.add_argument(
      '--federated-catalog-type',
      choices=choices,
      hidden_choices=hidden_choices,
      help='Type of the federated catalog.',
  )
  parser.add_argument(
      '--secret-name',
      help=(
          'The secret resource name in Secret Manager, in the format'
          ' `projects/{project_id}/locations/{location}/secrets/{secret_id}`'
          ' or `projects/{project_id}/locations/{location}/secrets/{secret_id}/'
          'versions/{version_id}`.'
      ),
  )

  parser.add_argument(
      '--unity-catalog-name',
      help='Name of the catalog in Unity Catalog.',
  )
  parser.add_argument(
      '--unity-instance-name',
      help=(
          'The instance name is the first part of the URL when you log into'
          ' your Databricks deployment. For example, for a Databricks on'
          ' Google Cloud workspace URL https://1.1.gcp.databricks.com, the'
          ' instance name is 1.1.gcp.databricks.com.'
      ),
  )
  parser.add_argument(
      '--refresh-interval',
      type=arg_parsers.Duration(),
      help=(
          'The interval for refreshing metadata from the remote catalog (e.g.,'
          ' "1hr5m30s"). Defaults to seconds if no unit is specified. If unset,'
          ' the refresh interval will be set to 0s (background refresh'
          ' disabled).'
      ),
  )
  parser.add_argument(
      '--namespace-filters',
      type=arg_parsers.ArgList(),
      metavar='NAMESPACE',
      help=(
          'Filters to determine which namespaces are included in the refresh'
          ' process. Empty list means include all namespaces.'
      ),
  )


def AddGlueCatalogArgs(parser):
  """Adds arguments for Glue catalogs."""
  parser.add_argument(
      '--glue-warehouse',
      help='The warehouse to connect to in AWS Glue Iceberg REST Catalog.',
  )
  parser.add_argument(
      '--glue-aws-region',
      help='The AWS region of the Glue catalog to connect to.',
  )
  AddGlueAwsRoleArnArg(parser)


def AddUpdateFederatedCatalogArgs(parser):
  """Adds arguments for updating federated catalogs."""
  parser.add_argument(
      '--secret-name',
      help=(
          'Resource name of the Secret Manager secret, in the format'
          ' `projects/{project_id}/locations/{location}/secrets/{secret_id}`'
          ' or `projects/{project_id}/locations/{location}/secrets/{secret_id}/'
          'versions/{version_id}`.'
      ),
  )

  parser.add_argument(
      '--refresh-interval',
      type=arg_parsers.Duration(),
      help=(
          'Interval for refreshing metadata from the remote catalog (e.g.,'
          ' "1hr5m30s"). Defaults to seconds if no unit is specified. If unset,'
          ' the refresh interval will be set to 0s (background refresh'
          ' disabled).'
      ),
  )
  parser.add_argument(
      '--namespace-filters',
      type=arg_parsers.ArgList(),
      metavar='NAMESPACE',
      help=(
          'Filters to determine which namespaces are included in the refresh'
          ' process. Empty list means include all namespaces.'
      ),
  )


def AddDefaultLocationArg(parser):
  """Adds argument for default location. Used for BigLake catalogs."""
  parser.add_argument(
      '--default-location',
      type=util.GcsBucketLinkValidator,
      help=(
          'Can only be used with BigLake catalogs. The default'
          ' storage location for the catalog, e.g., `gs://my-bucket/...`.'
      ),
  )


def AddRestrictedLocationsArg(parser):
  """Adds argument for restricted locations. Used for BigLake catalogs."""
  parser.add_argument(
      '--restricted-locations',
      type=arg_parsers.ArgList(element_type=util.GcsBucketLinkValidator),
      metavar='LOCATION',
      help=(
          'Additional Google Cloud Storage buckets and locations (e.g.,'
          ' `gs://my-other-bucket/...`) that are permitted for use by'
          ' resources within a catalog. This field is currently only used'
          ' for BigLake catalogs.'
          'If `restricted_locations` is empty and unrestricted catalog'
          ' creation is enabled, all accessible locations are allowed.'
          ' Otherwise, only `default_location` and locations in this'
          ' list are allowed.'
      ),
  )


def AddGlueAwsRoleArnArg(parser):
  """Adds Glue AWS role ARN argument."""
  parser.add_argument(
      '--glue-aws-role-arn',
      help=(
          'The AWS role ARN of the Glue catalog that the BigLake federated'
          ' catalog will assume to access the catalog.'
      ),
  )


def AddSnowflakeCatalogArgs(parser):
  """Adds arguments for Snowflake catalogs."""
  parser.add_argument(
      '--snowflake-warehouse',
      hidden=True,
      help='The warehouse to connect to in Snowflake.',
  )
  parser.add_argument(
      '--snowflake-account-identifier',
      hidden=True,
      help='The account identifier of the Snowflake catalog to connect to.',
  )


def AddWorkdayCatalogArgs(parser):
  """Adds arguments for Workday catalogs."""
  parser.add_argument(
      '--workday-base-url',
      hidden=True,
      help='The base URL of the Workday instance.',
  )
  parser.add_argument(
      '--workday-tenant',
      hidden=True,
      help='The Workday tenant name.',
  )


def AddHiveTableCreateArgs(parser):
  """Adds arguments for creating Hive tables."""
  parser.add_argument(
      '--file',
      required=True,
      help=(
          'Path to a JSON or YAML file containing the Hive table definition.'
          ' The format must follow the Hive Metastore API specification for'
          ' `HiveTable`. The field `storageDescriptor` must be specified.'
      ),
  )



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cloud Datastream connection profiles API."""


import json
from googlecloudsdk.api_lib.datastream import exceptions as ds_exceptions
from googlecloudsdk.api_lib.datastream import util
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.util.args import labels_util
from googlecloudsdk.core import resources
from googlecloudsdk.core import yaml
from googlecloudsdk.core.console import console_io

_DEFAULT_API_VERSION = 'v1'


def GetStreamURI(resource):
  stream = resources.REGISTRY.ParseRelativeName(
      resource.name,
      collection='datastream.projects.locations.streams')
  return stream.SelfLink()


class StreamsClient:
  """Client for streams service in the API."""

  def __init__(self, client=None, messages=None, location=None):
    self._client = client or util.GetClientInstance(location=location)
    self._messages = messages or util.GetMessagesModule()
    self._service = self._client.projects_locations_streams
    self._resource_parser = util.GetResourceParser()

  def _GetBackfillAllStrategy(self, release_track, args):
    """Gets BackfillAllStrategy message based on Stream objects source type."""
    if args.oracle_excluded_objects:
      return self._messages.BackfillAllStrategy(
          oracleExcludedObjects=util.ParseOracleRdbmsFile(
              self._messages, args.oracle_excluded_objects, release_track))
    elif args.mysql_excluded_objects:
      return self._messages.BackfillAllStrategy(
          mysqlExcludedObjects=util.ParseMysqlRdbmsFile(
              self._messages, args.mysql_excluded_objects, release_track))
    elif args.postgresql_excluded_objects:
      return self._messages.BackfillAllStrategy(
          postgresqlExcludedObjects=util.ParsePostgresqlRdbmsFile(
              self._messages, args.postgresql_excluded_objects))
    elif args.sqlserver_excluded_objects:
      return self._messages.BackfillAllStrategy(
          sqlServerExcludedObjects=util.ParseSqlServerRdbmsFile(
              self._messages, args.sqlserver_excluded_objects
          )
      )
    elif args.salesforce_excluded_objects:
      return self._messages.BackfillAllStrategy(
          salesforceExcludedObjects=util.ParseSalesforceOrgFile(
              self._messages, args.salesforce_excluded_objects
          )
      )
    elif args.spanner_excluded_objects:
      return self._messages.BackfillAllStrategy(
          spannerExcludedObjects=util.ParseSpannerDatabaseFile(
              self._messages, args.spanner_excluded_objects
          )
      )
    elif args.mongodb_excluded_objects:
      return self._messages.BackfillAllStrategy(
          mongodbExcludedObjects=util.ParseMongodbFile(
              self._messages, args.mongodb_excluded_objects
          )
      )
    elif args.IsSpecified('saas_excluded_objects'):
      return self._messages.BackfillAllStrategy(
          saasExcludedObjects=util.ParseSourceCatalogFile(
              self._messages, args.saas_excluded_objects
          )
      )
    return self._messages.BackfillAllStrategy()

  def _ParseOracleSourceConfig(self, oracle_source_config_file, release_track):
    """Parses a oracle_sorce_config into the OracleSourceConfig message."""
    if release_track == base.ReleaseTrack.BETA:
      return self._ParseOracleSourceConfigBeta(
          oracle_source_config_file, release_track
      )

    return util.ParseMessageAndValidateSchema(
        oracle_source_config_file,
        'OracleSourceConfig',
        self._messages.OracleSourceConfig,
    )

  def _ParseOracleSourceConfigBeta(
      self, oracle_source_config_file, release_track
  ):
    """Parses a oracle_sorce_config into the OracleSourceConfig message."""
    data = console_io.ReadFromFileOrStdin(
        oracle_source_config_file, binary=False)
    try:
      oracle_source_config_head_data = yaml.load(data)
    except yaml.YAMLParseError as e:
      raise ds_exceptions.ParseError('Cannot parse YAML:[{0}]'.format(e))

    oracle_sorce_config_data_object = oracle_source_config_head_data.get(
        'oracle_source_config'
    )
    oracle_source_config = (
        oracle_sorce_config_data_object
        if oracle_sorce_config_data_object
        else oracle_source_config_head_data
    )

    include_objects_raw = oracle_source_config.get(
        util.GetRDBMSV1alpha1ToV1FieldName('include_objects', release_track),
        {})
    include_objects_data = util.ParseOracleSchemasListToOracleRdbmsMessage(
        self._messages, include_objects_raw, release_track)

    exclude_objects_raw = oracle_source_config.get(
        util.GetRDBMSV1alpha1ToV1FieldName('exclude_objects', release_track),
        {})
    exclude_objects_data = util.ParseOracleSchemasListToOracleRdbmsMessage(
        self._messages, exclude_objects_raw, release_track)

    oracle_source_config_msg = self._messages.OracleSourceConfig(
        includeObjects=include_objects_data,
        excludeObjects=exclude_objects_data,
    )

    if oracle_source_config.get('max_concurrent_cdc_tasks'):
      oracle_source_config_msg.maxConcurrentCdcTasks = oracle_source_config.get(
          'max_concurrent_cdc_tasks')

    return oracle_source_config_msg

  def _ParseMysqlSourceConfig(self, mysql_source_config_file, release_track):
    """Parses a mysql_sorce_config into the MysqlSourceConfig message."""
    if release_track == base.ReleaseTrack.BETA:
      return self._ParseMysqlSourceConfigBeta(
          mysql_source_config_file, release_track
      )

    return util.ParseMessageAndValidateSchema(
        mysql_source_config_file,
        'MysqlSourceConfig',
        self._messages.MysqlSourceConfig,
    )

  def _ParseMysqlSourceConfigBeta(
      self, mysql_source_config_file, release_track
  ):
    """Parses an old mysql_sorce_config into the MysqlSourceConfig message."""
    data = console_io.ReadFromFileOrStdin(
        mysql_source_config_file, binary=False)
    try:
      mysql_sorce_config_head_data = yaml.load(data)
    except yaml.YAMLParseError as e:
      raise ds_exceptions.ParseError('Cannot parse YAML:[{0}]'.format(e))

    mysql_sorce_config_data_object = mysql_sorce_config_head_data.get(
        'mysql_source_config'
    )
    mysql_source_config = (
        mysql_sorce_config_data_object
        if mysql_sorce_config_data_object
        else mysql_sorce_config_head_data
    )

    include_objects_raw = mysql_source_config.get(
        util.GetRDBMSV1alpha1ToV1FieldName('include_objects', release_track),
        {})
    include_objects_data = util.ParseMysqlSchemasListToMysqlRdbmsMessage(
        self._messages, include_objects_raw, release_track)

    exclude_objects_raw = mysql_source_config.get(
        util.GetRDBMSV1alpha1ToV1FieldName('exclude_objects', release_track),
        {})
    exclude_objects_data = util.ParseMysqlSchemasListToMysqlRdbmsMessage(
        self._messages, exclude_objects_raw, release_track)

    mysql_sourec_config_msg = self._messages.MysqlSourceConfig(
        includeObjects=include_objects_data,
        excludeObjects=exclude_objects_data,
    )

    if mysql_source_config.get('max_concurrent_cdc_tasks'):
      mysql_sourec_config_msg.maxConcurrentCdcTasks = mysql_source_config.get(
          'max_concurrent_cdc_tasks')

    return mysql_sourec_config_msg

  def _ParsePostgresqlSourceConfig(self, postgresql_source_config_file):
    """Parses a postgresql_sorce_config into the PostgresqlSourceConfig message."""

    return util.ParseMessageAndValidateSchema(
        postgresql_source_config_file,
        'PostgresqlSourceConfig',
        self._messages.PostgresqlSourceConfig,
    )

  def _ParseSqlServerSourceConfig(self, sqlserver_source_config_file):
    """Parses a sqlserver_sorce_config into the SqlServerSourceConfig message."""

    return util.ParseMessageAndValidateSchema(
        sqlserver_source_config_file,
        'SqlServerSourceConfig',
        self._messages.SqlServerSourceConfig,
    )

  def _ParseSalesforceSourceConfig(self, salesforce_source_config_file):
    """Parses a salesforce_sorce_config into the SalesforceSourceConfig message."""

    return util.ParseMessageAndValidateSchema(
        salesforce_source_config_file,
        'SalesforceSourceConfig',
        self._messages.SalesforceSourceConfig,
    )

  def _ParseSpannerSourceConfig(self, spanner_source_config_file):
    """Parses a spanner_source_config into the SpannerSourceConfig message."""

    return util.ParseMessageAndValidateSchema(
        spanner_source_config_file,
        'SpannerSourceConfig',
        self._messages.SpannerSourceConfig,
    )

  def _ParseMongodbSourceConfig(self, mongodb_source_config_file):
    """Parses a mongodb_source_config into the MongodbSourceConfig message."""

    return util.ParseMessageAndValidateSchema(
        mongodb_source_config_file,
        'MongodbSourceConfig',
        self._messages.MongodbSourceConfig,
    )

  def _ParseDataverseSourceConfig(self, dataverse_source_config_file):
    """Parses a dataverse_source_config into the DataverseSourceConfig message."""
    return util.ParseMessageAndValidateSchema(
        dataverse_source_config_file,
        'DataverseSourceConfig',
        self._messages.DataverseSourceConfig,
    )

  def _ParseSalesforceMarketingCloudSourceConfig(
      self, salesforce_marketing_cloud_source_config_file
  ):
    """Parses a salesforce_marketing_cloud_source_config into the SalesforceMarketingCloudSourceConfig message."""
    return util.ParseMessageAndValidateSchema(
        salesforce_marketing_cloud_source_config_file,
        'SalesforceMarketingCloudSourceConfig',
        self._messages.SalesforceMarketingCloudSourceConfig,
    )

  def _ParseServiceNowSourceConfig(self, servicenow_source_config_file):
    """Parses a servicenow_source_config into the ServiceNowSourceConfig message."""
    return util.ParseMessageAndValidateSchema(
        servicenow_source_config_file,
        'ServiceNowSourceConfig',
        self._messages.ServiceNowSourceConfig,
    )

  def _ParseGcsDestinationConfig(
      self, gcs_destination_config_file, release_track
  ):
    """Parses a GcsDestinationConfig into the GcsDestinationConfig message."""

    if release_track == base.ReleaseTrack.BETA:
      return self._ParseGcsDestinationConfigBeta(gcs_destination_config_file)

    return util.ParseMessageAndValidateSchema(
        gcs_destination_config_file,
        'GcsDestinationConfig',
        self._messages.GcsDestinationConfig,
    )

  def _ParseGcsDestinationConfigBeta(self, gcs_destination_config_file):
    """Parses a gcs_destination_config into the GcsDestinationConfig message."""
    data = console_io.ReadFromFileOrStdin(
        gcs_destination_config_file, binary=False)
    try:
      gcs_destination_head_config_data = yaml.load(data)
    except yaml.YAMLParseError as e:
      raise ds_exceptions.ParseError('Cannot parse YAML:[{0}]'.format(e))

    gcs_destination_config_data_object = gcs_destination_head_config_data.get(
        'gcs_destination_config'
    )
    gcs_destination_config_data = (
        gcs_destination_config_data_object
        if gcs_destination_config_data_object
        else gcs_destination_head_config_data
    )

    path = gcs_destination_config_data.get('path', '')
    file_rotation_mb = gcs_destination_config_data.get('file_rotation_mb', {})
    file_rotation_interval = gcs_destination_config_data.get(
        'file_rotation_interval', {})
    gcs_dest_config_msg = self._messages.GcsDestinationConfig(
        path=path, fileRotationMb=file_rotation_mb,
        fileRotationInterval=file_rotation_interval)
    if 'avro_file_format' in gcs_destination_config_data:
      gcs_dest_config_msg.avroFileFormat = self._messages.AvroFileFormat()
    elif 'json_file_format' in gcs_destination_config_data:
      json_file_format_data = gcs_destination_config_data.get(
          'json_file_format')
      gcs_dest_config_msg.jsonFileFormat = self._messages.JsonFileFormat(
          compression=json_file_format_data.get('compression'),
          schemaFileFormat=json_file_format_data.get('schema_file_format'))
    else:
      raise ds_exceptions.ParseError(
          'Cannot parse YAML: missing file format.')
    return gcs_dest_config_msg

  def _ParseRuleSets(self, rule_sets_file):
    """Parses a list of RuleSets from a JSON file."""
    data = util.console_io.ReadFromFileOrStdin(rule_sets_file, binary=False)
    try:
      parsed_rule_sets = json.loads(data)
    except json.JSONDecodeError as e:
      raise ds_exceptions.ParseError('Cannot parse JSON:[{0}]'.format(e))

    rule_sets_list = []
    for rule_set in parsed_rule_sets:
      rule_sets_list.append(
          util.ParseJsonAndValidateSchema(
              rule_set, 'RuleSet', self._messages.RuleSet
          )
      )
    return rule_sets_list

  def _ParseBigqueryDestinationConfig(self, config_file):
    """Parses a BigQueryDestinationConfig into the BigQueryDestinationConfig message."""
    return util.ParseMessageAndValidateSchema(
        config_file,
        'BigQueryDestinationConfig',
        self._messages.BigQueryDestinationConfig,
    )

  def _GetStream(self, stream_id, release_track, args):
    """Returns a stream object."""
    labels = labels_util.ParseCreateArgs(
        args, self._messages.Stream.LabelsValue)
    stream_obj = self._messages.Stream(
        name=stream_id, labels=labels, displayName=args.display_name)

    # TODO(b/207467120): use CONCEPTS.source only.
    if release_track == base.ReleaseTrack.BETA:
      source_connection_profile_ref = args.CONCEPTS.source_name.Parse()
    else:
      source_connection_profile_ref = args.CONCEPTS.source.Parse()

    stream_source_config = self._messages.SourceConfig()
    stream_source_config.sourceConnectionProfile = (
        source_connection_profile_ref.RelativeName())
    if args.oracle_source_config:
      stream_source_config.oracleSourceConfig = self._ParseOracleSourceConfig(
          args.oracle_source_config, release_track)
    elif args.mysql_source_config:
      stream_source_config.mysqlSourceConfig = self._ParseMysqlSourceConfig(
          args.mysql_source_config, release_track)
    elif args.postgresql_source_config:
      stream_source_config.postgresqlSourceConfig = (
          self._ParsePostgresqlSourceConfig(args.postgresql_source_config)
      )
    elif args.sqlserver_source_config:
      stream_source_config.sqlServerSourceConfig = (
          self._ParseSqlServerSourceConfig(args.sqlserver_source_config)
      )
    elif args.salesforce_source_config:
      stream_source_config.salesforceSourceConfig = (
          self._ParseSalesforceSourceConfig(args.salesforce_source_config)
      )
    elif args.spanner_source_config:
      stream_source_config.spannerSourceConfig = (
          self._ParseSpannerSourceConfig(args.spanner_source_config)
      )
    elif args.mongodb_source_config:
      stream_source_config.mongodbSourceConfig = self._ParseMongodbSourceConfig(
          args.mongodb_source_config
      )
    elif args.dataverse_source_config:
      stream_source_config.dataverseSourceConfig = (
          self._ParseDataverseSourceConfig(args.dataverse_source_config)
      )
    elif args.salesforce_marketing_cloud_source_config:
      stream_source_config.salesforceMarketingCloudSourceConfig = (
          self._ParseSalesforceMarketingCloudSourceConfig(
              args.salesforce_marketing_cloud_source_config
          )
      )
    elif args.servicenow_source_config:
      stream_source_config.serviceNowSourceConfig = (
          self._ParseServiceNowSourceConfig(args.servicenow_source_config)
      )
    stream_obj.sourceConfig = stream_source_config

    # TODO(b/207467120): use CONCEPTS.destination only.
    if release_track == base.ReleaseTrack.BETA:
      destination_connection_profile_ref = args.CONCEPTS.destination_name.Parse(
      )
    else:
      destination_connection_profile_ref = args.CONCEPTS.destination.Parse()

    stream_destination_config = self._messages.DestinationConfig()
    stream_destination_config.destinationConnectionProfile = (
        destination_connection_profile_ref.RelativeName())
    if args.gcs_destination_config:
      stream_destination_config.gcsDestinationConfig = (
          self._ParseGcsDestinationConfig(
              args.gcs_destination_config, release_track
          )
      )
    elif args.bigquery_destination_config:
      stream_destination_config.bigqueryDestinationConfig = (
          self._ParseBigqueryDestinationConfig(
              args.bigquery_destination_config))
    stream_obj.destinationConfig = stream_destination_config

    if args.backfill_none:
      stream_obj.backfillNone = self._messages.BackfillNoneStrategy()
    elif args.backfill_all:
      backfill_all_strategy = self._GetBackfillAllStrategy(release_track, args)
      stream_obj.backfillAll = backfill_all_strategy

    if args.rule_sets:
      stream_obj.ruleSets = self._ParseRuleSets(args.rule_sets)

    return stream_obj

  def _GetExistingStream(self, name):
    get_req = self._messages.DatastreamProjectsLocationsStreamsGetRequest(
        name=name
    )
    return self._service.Get(get_req)

  def _UpdateLabels(self, stream, args):
    """Updates labels of the stream."""
    add_labels = labels_util.GetUpdateLabelsDictFromArgs(args)
    remove_labels = labels_util.GetRemoveLabelsListFromArgs(args)
    value_type = self._messages.Stream.LabelsValue
    update_result = labels_util.Diff(
        additions=add_labels,
        subtractions=remove_labels,
        clear=args.clear_labels
    ).Apply(value_type, stream.labels)
    if update_result.needs_update:
      stream.labels = update_result.labels

  def _UpdateListWithFieldNamePrefixes(
      self, update_fields, prefix_to_check, prefix_to_add):
    """Returns an updated list of field masks with necessary prefixes."""
    temp_fields = [
        prefix_to_add + field
        for field in update_fields
        if field.startswith(prefix_to_check)
    ]
    update_fields = [
        x for x in update_fields if (not x.startswith(prefix_to_check))
    ]
    update_fields.extend(temp_fields)
    return update_fields

  def _GetUpdatedStream(self, stream, release_track, args):
    """Returns updated stream."""
    # Verify command flag names align with Stream object field names.
    update_fields = []
    user_update_mask = args.update_mask or ''
    user_update_mask_list = user_update_mask.split(',')

    if release_track == base.ReleaseTrack.BETA:
      user_update_mask_list = util.UpdateV1alpha1ToV1MaskFields(
          user_update_mask_list)

    update_fields.extend(user_update_mask_list)

    if args.IsSpecified('display_name'):
      stream.displayName = args.display_name

    # TODO(b/207467120): use source field only.
    if release_track == base.ReleaseTrack.BETA:
      source_connection_profile_ref = args.CONCEPTS.source_name.Parse()
      source_field_name = 'source_name'
    else:
      source_connection_profile_ref = args.CONCEPTS.source.Parse()
      source_field_name = 'source'
    if args.IsSpecified(source_field_name):
      stream.sourceConfig.sourceConnectionProfile = (
          source_connection_profile_ref.RelativeName())
      if source_field_name in update_fields:
        update_fields.remove(source_field_name)
        update_fields.append('source_config.source_connection_profile')

    if args.IsSpecified('oracle_source_config'):
      stream.sourceConfig.oracleSourceConfig = self._ParseOracleSourceConfig(
          args.oracle_source_config, release_track)
      # Fix field names in update mask
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'oracle_source_config', 'source_config.')

    elif args.IsSpecified('mysql_source_config'):
      stream.sourceConfig.mysqlSourceConfig = self._ParseMysqlSourceConfig(
          args.mysql_source_config, release_track)
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'mysql_source_config', 'source_config.')

    elif args.IsSpecified('postgresql_source_config'):
      stream.sourceConfig.postgresqlSourceConfig = (
          self._ParsePostgresqlSourceConfig(args.postgresql_source_config)
      )
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'postgresql_source_config', 'source_config.')

    elif args.IsSpecified('sqlserver_source_config'):
      stream.sourceConfig.sqlServerSourceConfig = (
          self._ParseSqlServerSourceConfig(args.sqlserver_source_config)
      )
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'sqlserver_source_config', 'source_config.'
      )
    elif args.IsSpecified('salesforce_source_config'):
      stream.sourceConfig.salesforceSourceConfig = (
          self._ParseSalesforceSourceConfig(args.salesforce_source_config)
      )
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'salesforce_source_config', 'source_config.'
      )
    elif args.IsSpecified('spanner_source_config'):
      stream.sourceConfig.spannerSourceConfig = (
          self._ParseSpannerSourceConfig(args.spanner_source_config)
      )
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'spanner_source_config', 'source_config.'
      )
    elif args.IsSpecified('dataverse_source_config'):
      stream.sourceConfig.dataverseSourceConfig = (
          self._ParseDataverseSourceConfig(args.dataverse_source_config)
      )
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'dataverse_source_config', 'source_config.'
      )
    elif args.IsSpecified('salesforce_marketing_cloud_source_config'):
      stream.sourceConfig.salesforceMarketingCloudSourceConfig = (
          self._ParseSalesforceMarketingCloudSourceConfig(
              args.salesforce_marketing_cloud_source_config
          )
      )
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields,
          'salesforce_marketing_cloud_source_config',
          'source_config.',
      )
    elif args.IsSpecified('servicenow_source_config'):
      stream.sourceConfig.serviceNowSourceConfig = (
          self._ParseServiceNowSourceConfig(args.servicenow_source_config)
      )
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'servicenow_source_config', 'source_config.'
      )

    # TODO(b/207467120): use source field only.
    if release_track == base.ReleaseTrack.BETA:
      destination_connection_profile_ref = (
          args.CONCEPTS.destination_name.Parse())
      destination_field_name = 'destination_name'
    else:
      destination_connection_profile_ref = (args.CONCEPTS.destination.Parse())
      destination_field_name = 'destination'
    if args.IsSpecified(destination_field_name):
      stream.destinationConfig.destinationConnectionProfile = (
          destination_connection_profile_ref.RelativeName())
      if destination_field_name in update_fields:
        update_fields.remove(destination_field_name)
        update_fields.append(
            'destination_config.destination_connection_profile')

    if args.IsSpecified('gcs_destination_config'):
      stream.destinationConfig.gcsDestinationConfig = (
          self._ParseGcsDestinationConfig(
              args.gcs_destination_config, release_track
          )
      )
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'gcs_destination_config', 'destination_config.')
    elif args.IsSpecified('bigquery_destination_config'):
      stream.destinationConfig.bigqueryDestinationConfig = (
          self._ParseBigqueryDestinationConfig(
              args.bigquery_destination_config))
      update_fields = self._UpdateListWithFieldNamePrefixes(
          update_fields, 'bigquery_destination_config', 'destination_config.')

    if args.IsSpecified('backfill_none'):
      stream.backfillNone = self._messages.BackfillNoneStrategy()
      # NOMUTANTS--This path has been verified by manual tests.
      try:
        stream.reset('backfillAll')
      except AttributeError:
        # Attempt to remove a backfill all
        # previous definition, but doesn't exist.
        pass

    elif args.IsSpecified('backfill_all'):
      backfill_all_strategy = self._GetBackfillAllStrategy(release_track, args)
      stream.backfillAll = backfill_all_strategy
      # NOMUTANTS--This path has been verified by manual tests.
      try:
        stream.reset('backfillNone')
      except AttributeError:
        # Attempt to remove a backfill none previous definition,
        # but it doesn't exist.
        pass

    if args.IsSpecified('state'):
      stream.state = self._messages.Stream.StateValueValuesEnum(
          (args.state).upper())

    if args.IsSpecified('rule_sets'):
      stream.ruleSets = self._ParseRuleSets(args.rule_sets)

    self._UpdateLabels(stream, args)
    return stream, update_fields

  def Create(self, parent_ref, stream_id, release_track, args=None):
    """Creates a stream.

    Args:
      parent_ref: a Resource reference to a parent datastream.projects.locations
        resource for this stream.
      stream_id: str, the name of the resource to create.
      release_track: Some arguments are added based on the command release
        track.
      args: argparse.Namespace, The arguments that this command was invoked
        with.

    Returns:
      Operation: the operation for creating the stream.
    """
    stream = self._GetStream(stream_id, release_track, args)
    validate_only = args.validate_only
    force = args.force

    request_id = util.GenerateRequestId()
    create_req_type = self._messages.DatastreamProjectsLocationsStreamsCreateRequest
    create_req = create_req_type(
        stream=stream,
        streamId=stream.name,
        parent=parent_ref,
        requestId=request_id,
        validateOnly=validate_only,
        force=force)

    return self._service.Create(create_req)

  def Update(self, name, release_track, args=None):
    """Updates a stream.

    Args:
      name: str, the reference of the stream to
          update.
      release_track: Some arguments are added based on the command release
        track.
      args: argparse.Namespace, The arguments that this command was
          invoked with.

    Returns:
      Operation: the operation for updating the connection profile.
    """
    validate_only = args.validate_only
    force = args.force

    current_stream = self._GetExistingStream(name)

    updated_stream, update_fields = self._GetUpdatedStream(
        current_stream, release_track, args)

    request_id = util.GenerateRequestId()
    update_req_type = self._messages.DatastreamProjectsLocationsStreamsPatchRequest
    update_req = update_req_type(
        stream=updated_stream,
        name=updated_stream.name,
        requestId=request_id,
        validateOnly=validate_only,
        force=force
    )
    if args.update_mask:
      update_req.updateMask = ','.join(update_fields)

    return self._service.Patch(update_req)



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Designcenter Locations API."""


from apitools.base.py import list_pager
from googlecloudsdk.api_lib.design_center import utils as api_lib_utils
from googlecloudsdk.calliope import base


class LocationsClient(object):
  """Client for locations in Designcenter API."""

  def __init__(self, release_track=base.ReleaseTrack.ALPHA):
    self.client = api_lib_utils.GetClientInstance(release_track)
    self.messages = api_lib_utils.GetMessagesModule(release_track)
    self._lo_client = self.client.projects_locations

  def List(self, parent, limit=None, page_size=100):
    """List all Designcenter locations in the Project.

    Args:
      parent: str, the project in which the locations being listed.
        projects/{projectId}
      limit: int or None, the total number of results to return.
        Default value is None
      page_size: int, the number of entries in each batch (affects requests
        made, but not the yielded results). Default value is 100.

    Returns:
      A list of Designcenter locations that belong to the given parent.
    """
    list_req = self.messages.DesigncenterProjectsLocationsListRequest(
        name=parent)
    return list_pager.YieldFromList(
        self._lo_client,
        list_req,
        field='locations',
        batch_size=page_size,
        limit=limit,
        batch_size_attribute='pageSize',
    )

  def Describe(self, location):
    """Describe a Designcenter location.

    Args:
      location: str, the name for the Designcenter Location being described.

    Returns:
      Described Designcenter location resource.
    """
    describe_req = self.messages.DesigncenterProjectsLocationsGetRequest(
        name=location)
    return self._lo_client.Get(describe_req)

  def FetchFrameworks(self, parent, projects=None):
    """Fetch frameworks available for assessment.

    Args:
      parent: str, the location in which frameworks are being fetched.
        projects/{projectId}/locations/{location}
      projects: list[str], the projects to fetch frameworks for.

    Returns:
      FetchFrameworksResponse message.
    """
    request = self.messages.DesigncenterProjectsLocationsFetchFrameworksRequest(
        parent=parent)
    if projects:
      request.projects = [
          p if p.startswith('projects/') else f'projects/{p}'
          for p in projects
      ]
    return self._lo_client.FetchFrameworks(request)



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common utility functions to consturct compute future reservations message."""


from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.compute import resource_manager_tags_utils
from googlecloudsdk.command_lib.compute.reservations import util as reservation_util
from googlecloudsdk.core.util import times


def MakeFutureReservationMessageFromArgs(messages, resources, args,
                                         future_reservation_ref):
  """Construct future reservation message from args passed in."""
  local_ssds = reservation_util.MakeLocalSsds(messages,
                                              getattr(args, 'local_ssd', None))
  accelerators = reservation_util.MakeGuestAccelerators(
      messages, getattr(args, 'accelerator', None))

  allocated_instance_properties = MakeAllocatedInstanceProperties(
      messages,
      args.machine_type,
      args.min_cpu_platform,
      local_ssds,
      accelerators,
      getattr(args, 'location_hint', None),
      getattr(args, 'maintenance_freeze_duration', None),
      getattr(args, 'maintenance_interval', None),
  )

  source_instance_template_ref = (
      reservation_util.ResolveSourceInstanceTemplate(args, resources)
      if getattr(args, 'source_instance_template', None)
      else None
  )

  sku_properties = MakeSpecificSKUPropertiesMessage(
      messages,
      allocated_instance_properties,
      args.total_count,
      source_instance_template_ref,
  )

  aggregate_reservation_properties = MakeAggregateReservationPropertiesMessage(
      messages,
      getattr(args, 'tpu_version', None),
      getattr(args, 'chip_count', None),
      getattr(args, 'workload_type', None),
  )

  time_window = MakeTimeWindowMessage(messages, args.start_time,
                                      getattr(args, 'end_time', None),
                                      getattr(args, 'duration', None))
  share_settings = MakeShareSettings(messages, args,
                                     getattr(args, 'share_setting', None))
  planning_status = MakePlanningStatus(messages,
                                       getattr(args, 'planning_status', None))

  enable_auto_delete_reservations = None
  if args.IsSpecified('auto_delete_auto_created_reservations'):
    enable_auto_delete_reservations = getattr(
        args, 'auto_delete_auto_created_reservations'
    )

  auto_created_reservations_delete_time = None
  if args.IsSpecified('auto_created_reservations_delete_time'):
    auto_created_reservations_delete_time = getattr(
        args, 'auto_created_reservations_delete_time'
    )
  auto_created_reservations_duration = None
  if args.IsSpecified('auto_created_reservations_duration'):
    auto_created_reservations_duration = getattr(
        args, 'auto_created_reservations_duration'
    )

  require_specific_reservation = getattr(
      args, 'require_specific_reservation', None
  )

  reservation_name = getattr(args, 'reservation_name', None)

  deployment_type = None
  if args.IsKnownAndSpecified('deployment_type'):
    deployment_type = MakeDeploymentType(messages,
                                         getattr(args, 'deployment_type', None))

  commitment_info = MakeCommitmentInfo(messages, args)
  scheduling_type = None
  if args.IsKnownAndSpecified('scheduling_type'):
    scheduling_type = MakeSchedulingType(
        messages, getattr(args, 'scheduling_type', None)
    )

  reservation_mode = None
  if args.IsKnownAndSpecified('reservation_mode'):
    reservation_mode = MakeReservationMode(
        messages, getattr(args, 'reservation_mode', None)
    )

  enable_emergent_maintenance = None
  if args.IsKnownAndSpecified('enable_emergent_maintenance'):
    enable_emergent_maintenance = getattr(
        args, 'enable_emergent_maintenance', None
    )

  resource_manager_tags = None
  if args.IsKnownAndSpecified('resource_manager_tags'):
    resource_manager_tags = MakeResourceManagerTags(
        messages, getattr(args, 'resource_manager_tags', None)
    )

  return MakeFutureReservationMessage(
      messages,
      future_reservation_ref.Name(),
      time_window,
      share_settings,
      planning_status,
      aggregate_reservation_properties,
      sku_properties,
      enable_auto_delete_reservations,
      auto_created_reservations_delete_time,
      auto_created_reservations_duration,
      require_specific_reservation,
      reservation_name,
      deployment_type,
      commitment_info,
      scheduling_type,
      reservation_mode,
      enable_emergent_maintenance,
      resource_manager_tags,
  )


def MakeAllocatedInstanceProperties(messages,
                                    machine_type,
                                    min_cpu_platform,
                                    local_ssds,
                                    accelerators,
                                    location_hint=None,
                                    freeze_duration=None,
                                    freeze_interval=None):
  """Constructs an instance propteries for reservations message object."""
  if machine_type is None:
    return None

  prop_msgs = (
      messages.AllocationSpecificSKUAllocationReservedInstanceProperties)
  instance_properties = prop_msgs(
      machineType=machine_type,
      guestAccelerators=accelerators,
      minCpuPlatform=min_cpu_platform,
      localSsds=local_ssds)
  if location_hint:
    instance_properties.locationHint = location_hint
  if freeze_duration:
    instance_properties.maintenanceFreezeDurationHours = freeze_duration // 3600
  if freeze_interval:
    instance_properties.maintenanceInterval = (
        messages.AllocationSpecificSKUAllocationReservedInstanceProperties
        .MaintenanceIntervalValueValuesEnum(freeze_interval))
  return instance_properties


def MakeSpecificSKUPropertiesMessage(
    messages,
    instance_properties,
    total_count,
    source_instance_template_ref=None,
):
  """Constructs a specific sku properties message object if any is specified."""
  if instance_properties is None and source_instance_template_ref is None:
    return None

  properties = None
  source_instance_template_url = None
  if source_instance_template_ref:
    source_instance_template_url = source_instance_template_ref.SelfLink()
  else:
    properties = instance_properties
  return messages.FutureReservationSpecificSKUProperties(
      totalCount=total_count,
      sourceInstanceTemplate=source_instance_template_url,
      instanceProperties=properties)


def MakeAggregateReservationPropertiesMessage(
    messages, tpu_version, chip_count, workload_type
):
  """Constructs an aggregate reservation properties message object."""
  if not tpu_version:
    return None

  # Static mapping of TPU version to VM family.
  tpu_version_to_vm_family = {
      'V5E': (
          messages.AllocationAggregateReservation.VmFamilyValueValuesEnum.VM_FAMILY_CLOUD_TPU_LITE_POD_SLICE_CT5LP
      ),
      'V5P': (
          messages.AllocationAggregateReservation.VmFamilyValueValuesEnum.VM_FAMILY_CLOUD_TPU_POD_SLICE_CT5P
      ),
      'V6E': (
          messages.AllocationAggregateReservation.VmFamilyValueValuesEnum.VM_FAMILY_CLOUD_TPU_LITE_POD_SLICE_CT6E
      ),
      'TPU7X': (
          messages.AllocationAggregateReservation.VmFamilyValueValuesEnum.VM_FAMILY_CLOUD_TPU_POD_SLICE_TPU7X
      ),
  }

  reserved_resources = []
  accelerator = (
      messages.AllocationAggregateReservationReservedResourceInfoAccelerator(
          acceleratorCount=chip_count,
      )
  )
  reserved_resources.append(
      messages.AllocationAggregateReservationReservedResourceInfo(
          accelerator=accelerator
      )
  )
  aggregate_reservation_properties = messages.AllocationAggregateReservation(
      vmFamily=messages.AllocationAggregateReservation.VmFamilyValueValuesEnum(
          tpu_version_to_vm_family[tpu_version]
      ),
      reservedResources=reserved_resources,
  )

  if workload_type:
    aggregate_reservation_properties.workloadType = (
        messages.AllocationAggregateReservation.WorkloadTypeValueValuesEnum(
            workload_type
        )
    )
  return aggregate_reservation_properties


def MakeTimeWindowMessage(messages, start_time, end_time, duration):
  """Constructs the time window message object."""
  if end_time:
    return messages.FutureReservationTimeWindow(
        startTime=start_time, endTime=end_time)
  else:
    return messages.FutureReservationTimeWindow(
        startTime=start_time, duration=messages.Duration(seconds=duration))


def MakeShareSettings(messages, args, setting_configs):
  """Constructs the share settings message object."""
  if setting_configs:
    if setting_configs == 'local':
      if args.IsSpecified('share_with'):
        raise exceptions.InvalidArgumentException(
            '--share_with',
            'The scope this reservation is to be shared with must not be '
            'specified with share setting local.')
      return messages.ShareSettings(shareType=messages.ShareSettings
                                    .ShareTypeValueValuesEnum.LOCAL)
    if setting_configs == 'projects':
      if not args.IsSpecified('share_with'):
        raise exceptions.InvalidArgumentException(
            '--share_with',
            'The projects this reservation is to be shared with must be '
            'specified.')
      return messages.ShareSettings(
          shareType=messages.ShareSettings.ShareTypeValueValuesEnum
          .SPECIFIC_PROJECTS,
          projectMap=MakeProjectMapFromProjectList(
              messages, getattr(args, 'share_with', None)))
    if setting_configs == 'folders':
      return messages.ShareSettings(
          shareType=messages.ShareSettings.ShareTypeValueValuesEnum
          .DIRECT_PROJECTS_UNDER_SPECIFIC_FOLDERS,
          folderMap=MakeFolderMapFromFolderList(
              messages, getattr(args, 'share_with', None)))
  else:
    return None


def MakeProjectMapFromProjectList(messages, projects):
  additional_properties = []
  for project in projects:
    additional_properties.append(
        messages.ShareSettings.ProjectMapValue.AdditionalProperty(
            key=project,
            value=messages.ShareSettingsProjectConfig(projectId=project)))
  return messages.ShareSettings.ProjectMapValue(
      additionalProperties=additional_properties)


def MakeFolderMapFromFolderList(messages, folders):
  additional_properties = []
  if folders:
    for folder in folders:
      additional_properties.append(
          messages.ShareSettings.FolderMapValue.AdditionalProperty(
              key=folder,
              value=messages.ShareSettingsFolderConfig(folderId=folder),
          )
      )
  return messages.ShareSettings.FolderMapValue(
      additionalProperties=additional_properties
  )


def MakePlanningStatus(messages, planning_status):
  """Constructs the planning status enum value."""
  if planning_status:
    if planning_status == 'SUBMITTED':
      return messages.FutureReservation.PlanningStatusValueValuesEnum.SUBMITTED
    if planning_status == 'DRAFT':
      return messages.FutureReservation.PlanningStatusValueValuesEnum.DRAFT
  return None


def MakeDeploymentType(messages, deployment_type):
  """Constructs the deployment type enum value."""
  if deployment_type:
    if deployment_type == 'DENSE':
      return messages.FutureReservation.DeploymentTypeValueValuesEnum.DENSE
    if deployment_type == 'FLEXIBLE':
      return messages.FutureReservation.DeploymentTypeValueValuesEnum.FLEXIBLE
  return None


def MakeCommitmentInfo(messages, args):
  """Constructs the commitment info message object."""
  commitment_name = getattr(args, 'commitment_name', None)
  commitment_plan = None
  if args.IsKnownAndSpecified('commitment_plan'):
    commitment_plan = MakeCommitmentPlan(messages,
                                         getattr(args, 'commitment_plan', None))
  previous_commitment_terms = None
  if args.IsKnownAndSpecified('previous_commitment_terms'):
    previous_commitment_terms = MakePreviousCommitmentTerms(
        messages, getattr(args, 'previous_commitment_terms', None)
    )

  if any([commitment_name, commitment_plan, previous_commitment_terms]):
    return messages.FutureReservationCommitmentInfo(
        commitmentName=commitment_name,
        commitmentPlan=commitment_plan,
        previousCommitmentTerms=previous_commitment_terms,
    )
  return None


def MakeCommitmentPlan(messages, commitment_plan):
  """Constructs the commitment plan enum value."""
  if commitment_plan:
    if commitment_plan == 'TWELVE_MONTH':
      return (messages.FutureReservationCommitmentInfo.CommitmentPlanValueValuesEnum.
              TWELVE_MONTH)
    if commitment_plan == 'THIRTY_SIX_MONTH':
      return (messages.FutureReservationCommitmentInfo.CommitmentPlanValueValuesEnum
              .THIRTY_SIX_MONTH)
  return None


def MakePreviousCommitmentTerms(messages, previous_commitment_terms):
  """Constructs the previous commitment terms enum value."""
  if previous_commitment_terms:
    if previous_commitment_terms == 'EXTEND':
      return (messages.FutureReservationCommitmentInfo.PreviousCommitmentTermsValueValuesEnum
              .EXTEND)
  return None


def MakeSchedulingType(messages, scheduling_type):
  """Constructs the scheduling type enum value."""
  if scheduling_type:
    if scheduling_type == 'GROUPED':
      return messages.FutureReservation.SchedulingTypeValueValuesEnum.GROUPED
    if scheduling_type == 'INDEPENDENT':
      return (messages.FutureReservation.SchedulingTypeValueValuesEnum
              .INDEPENDENT)
  return None


def MakeReservationMode(messages, reservation_mode):
  """Constructs the reservation mode enum value."""
  if reservation_mode:
    if reservation_mode == 'CALENDAR':
      return messages.FutureReservation.ReservationModeValueValuesEnum.CALENDAR
    if reservation_mode == 'DEFAULT':
      return messages.FutureReservation.ReservationModeValueValuesEnum.DEFAULT
  return None


def MakeResourceManagerTags(messages, resource_manager_tags):
  """Creates a future reservation params object for resource manager tags."""
  if resource_manager_tags is None:
    return None
  resource_manager_tags_map = (
      resource_manager_tags_utils.GetResourceManagerTags(resource_manager_tags)
  )
  params = messages.FutureReservationParams
  additional_properties = [
      params.ResourceManagerTagsValue.AdditionalProperty(key=key, value=value)
      for key, value in sorted(resource_manager_tags_map.items())
  ]
  return params(
      resourceManagerTags=params.ResourceManagerTagsValue(
          additionalProperties=additional_properties
      )
  )


def MakeFutureReservationMessage(
    messages,
    future_reservation_name,
    time_window,
    share_settings,
    planning_status,
    aggregate_reservation_properties=None,
    sku_properties=None,
    enable_auto_delete_reservations=None,
    auto_created_reservations_delete_time=None,
    auto_created_reservations_duration=None,
    require_specific_reservation=None,
    reservation_name=None,
    deployment_type=None,
    commitment_info=None,
    scheduling_type=None,
    reservation_mode=None,
    enable_emergent_maintenance=None,
    resource_manager_tags=None,
):
  """Constructs a future reservation message object."""
  future_reservation_message = messages.FutureReservation(
      name=future_reservation_name,
      timeWindow=time_window,
      planningStatus=planning_status,
  )

  if aggregate_reservation_properties:
    future_reservation_message.aggregateReservation = (
        aggregate_reservation_properties
    )
  if sku_properties:
    future_reservation_message.specificSkuProperties = sku_properties

  if share_settings:
    future_reservation_message.shareSettings = share_settings

  if enable_auto_delete_reservations is not None:
    future_reservation_message.autoDeleteAutoCreatedReservations = (
        enable_auto_delete_reservations
    )

  if auto_created_reservations_delete_time is not None:
    future_reservation_message.autoCreatedReservationsDeleteTime = (
        times.FormatDateTime(auto_created_reservations_delete_time)
    )
  if auto_created_reservations_duration is not None:
    future_reservation_message.autoCreatedReservationsDuration = (
        messages.Duration(seconds=auto_created_reservations_duration)
    )
  if require_specific_reservation is not None:
    future_reservation_message.specificReservationRequired = (
        require_specific_reservation
    )

  if reservation_name is not None:
    future_reservation_message.reservationName = reservation_name
  if deployment_type is not None:
    future_reservation_message.deploymentType = deployment_type
  if commitment_info is not None:
    future_reservation_message.commitmentInfo = commitment_info
  if scheduling_type is not None:
    future_reservation_message.schedulingType = scheduling_type
  if reservation_mode is not None:
    future_reservation_message.reservationMode = reservation_mode
  if enable_emergent_maintenance is not None:
    future_reservation_message.enableEmergentMaintenance = (
        enable_emergent_maintenance
    )
  if resource_manager_tags is not None:
    future_reservation_message.params = resource_manager_tags

  return future_reservation_message
